# -*- coding: utf-8 -*-

"""
提供主要用于数据交换的序列化处理机制。
"""

__all__ = ['Serializer']

import datetime
import decimal
import json
import logging
import six

try:
    import cPickle as pickle
except ImportError:
    import pickle
    logging.warning("can't import cpickle, use pickle instead")


class ObjectDict(dict):
    """Makes a dictionary behave like an object."""

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        del self[name]


class Serializer(object):
    """序列化处理器"""

    #: 支持的序列化格式。
    SUPPORTED_FORMATS = ['JSON', 'PICKLE']

    def __init__(self, format='JSON'):
        """创建一个序列化处理器。

        :param str format: 指定该序列化处理器采用的格式，如 YAML、JSON 等。
        """
        format = format.upper()
        if format in self.SUPPORTED_FORMATS:
            self.format = format
        else:
            raise ValueError('unsupported serializaion format')

    def load(self, stream):
        """从参数 ``stream`` 中获取数据。

        :param stream: 要载入数据的来源。可以是字符串或文件等类型。
        :type stream: mixed
        :rtype: str|unicode|file
        """
        func_name = ''.join(['_from_', self.format.lower()])
        func = globals()[func_name]
        return func(stream)

    def dump(self, data):
        """将指定数据 ``data`` 转换为序列化后的信息。

        :param data: 要序列化的数据。通常是某种映射或序列类型。
        :type data: mixed
        :rtype: str|unicode
        """
        func_name = ''.join(['_to_', self.format.lower()])
        func = globals()[func_name]
        return func(data)

    def serialize(self, data):
        """:meth:`dump` 方法的别名。"""
        return self.dump(data)

    def unserialize(self, stream):
        """:meth:`load` 方法的别名。"""
        return self.load(stream)

    def encode(self, data):
        """:meth:`dump` 方法的别名。"""
        return self.dump(data)

    def decode(self, stream):
        """:meth:`load` 方法的别名。"""
        return self.load(stream)


def _from_pickle(stream):
    """Load data from PICKLE file or string"""
    if isinstance(stream, six.string_types):
        data = pickle.loads(stream)
    else:
        data = pickle.load(stream)
    return data


def _to_pickle(data):
    """Dump data into a PICKLE string."""
    return pickle.dumps(data)


def _from_json(stream):
    """Load data form a JSON file or string."""
    if isinstance(stream, six.string_types) or isinstance(stream, six.binary_type):
        data = json.loads(stream, object_hook=lambda d: ObjectDict(d))
    else:
        data = json.load(stream, object_hook=lambda d: ObjectDict(d))
    return data


def _to_json(data):
    """Dump data into a JSON string."""
    return json.dumps(data, cls=AwareJSONEncoder)


class AwareJSONEncoder(json.JSONEncoder):
    """JSONEncoder subclass that knows how to encode date/time and
    decimal types, and also ResultProxy/RowProxy of SQLAlchemy.
    """

    DATE_FORMAT = "%Y-%m-%d"
    TIME_FORMAT = "%H:%M:%S"

    def default(self, o):
        from sqlalchemy.engine import ResultProxy, RowProxy
        from sqlalchemy.ext.declarative import DeclarativeMeta
        from sqlalchemy.util._collections import AbstractKeyedTuple
        if isinstance(o, datetime.datetime):
            # d = datetime_safe.new_datetime(o)
            # return d.strftime("%s %s" % (self.DATE_FORMAT, self.TIME_FORMAT))
            return o.strftime("%s %s" % (self.DATE_FORMAT, self.TIME_FORMAT))
        elif isinstance(o, datetime.date):
            # d = datetime_safe.new_date(o)
            return o.strftime(self.DATE_FORMAT)
        elif isinstance(o, datetime.time):
            return o.strftime(self.TIME_FORMAT)
        elif isinstance(o, decimal.Decimal):
            return str(o)
        elif isinstance(o, ResultProxy):
            return list(o)
        elif isinstance(o, RowProxy):
            return dict(o)
        elif isinstance(o, AbstractKeyedTuple):
            return o._asdict()
        elif isinstance(o.__class__, DeclarativeMeta):
            fields = {}
            instance_dict = o.__dict__
            for field in instance_dict:
                if not field.startswith('_'):
                    fields[field] = instance_dict[field]
            return fields
        else:
            return super(AwareJSONEncoder, self).default(o)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
DEFAULT_TIMEOUT = 3600

class NotFound(object):
    """Describe a missing cache value.

    The :attr:`.NO_VALUE` module global
    should be used.
    """
    __slots__ = tuple()

    def __repr__(self):
        return 'NotFound'


NOT_FOUND = NotFound()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Define a store class, which  contains the `get` and `set` method
>>> from .store import RedisStore
>>> import fakeredis
>>> redis_client = fakeredis.FakeStrictRedis()

Construct a cache instance
>>> cache = CacheFactory.from_store_conn("redis", redis_client)

Use the cache on function

>>> @cache("add:{a}:{b}")
... def add(a, b, c=1):
...    return a + b + c
>>> add(3, 2)
6

The third argument c is not used in the cache key "add:{a}:{b}",
so the result will not change.
>>> add(3, b=3)
7
>>> add(3, b=3, c=2)
7
>>> add.invalidate(3, 3)
>>> add(3, b=3, c=2)
8

You can use positional argument or keywords argument as you wish.
>>> @cache("mul:{a}:{b}:{c}")
... def mul(a, b, c=2):
...   return a * b * c
>>> mul(3, 4)
24
>>> mul(3, 4) == mul(3, c=2, b=4) == mul(c=2, a=3, b=4) == mul(3, 4, 2)
True

Use the cache on class

>>> class Example(object):
...   def __init__(self, incr):
...       self.incr = incr
...   @cache("example.add:{a}:{b}:{c}")
...   def add(self, a, b, c=1):
...       return a + b + c + self.incr
...   @cache("example.mul:{a}:{b}:{c}")
...   @classmethod
...   def mul(cls, a, b, c=2):
...       return a * b * c

>>> Example(1).add(2, 3)
7
>>> Example(2).add(a=2, b=3) == Example(3).add(b=3, c=1, a=2)
True
>>> Example.mul(3,4)
24
>>> Example.mul(3,4) == Example.mul(3, b=4, c=2) == Example.mul(b=4,c=2,a=3)
True

>>> @cache("add_ref:{a.id}:{b.id}")
... def add_ref(a, b, c=1):
...    return a.id + b.id + c
>>> from collections import namedtuple
>>> Point = namedtuple('Point', 'id')
>>> add_ref(Point(3), Point(4))
8
'''
import inspect
import types

from .store import RedisStore
from .constants import NOT_FOUND, DEFAULT_TIMEOUT


class Cache(object):

    def __init__(self, store, prefix, func, ruler, timeout):
        self._store = store
        self._prefix = prefix
        self._func = func
        self._ruler = ruler
        self._timeout = timeout

    def __get__(self, instance, owner):
        wrapped_self = object.__new__(self.__class__)
        wrapped_self.__dict__ = self.__dict__.copy()
        if instance is None:
            if not hasattr(self._func, "__call__"):
                wrapped_self._func = self._func.__get__(None, owner)
            return wrapped_self
        if not isinstance(self._func, types.MethodType):
            wrapped_self._func = self._func.__get__(instance, owner)
        return wrapped_self

    def __call__(self, *args, **kwargs):
        kwargs = self.__prune_arguments(*args, **kwargs)
        cache_key = self._ruler.format(**kwargs)
        if self._prefix:
            cache_key = self._prefix + cache_key
        result = self._store.get(cache_key)
        if result is NOT_FOUND:
            result = self._func(**kwargs)
            self._store.set(cache_key, result, self._timeout)
        return result

    def __prune_arguments(self, *args, **kwargs):
        arg_spec = inspect.getargspec(self._func)
        arg_names = arg_spec.args
        defaults = arg_spec.defaults
        if args:
            if isinstance(self._func, types.MethodType):
                kwargs.update(dict(zip(arg_names[1:], args)))
            else:
                kwargs.update(dict(zip(arg_names, args)))
        if defaults:
            for pos, value in enumerate(defaults):
                argname = arg_names[-len(defaults) + pos]
                if argname not in kwargs:
                    kwargs[argname] = value
        return kwargs

    def invalidate(self, *args, **kwargs):
        kwargs = self.__prune_arguments(*args, **kwargs)
        cache_key = self._ruler.format(**kwargs)
        if self._prefix:
            cache_key = self._prefix + cache_key
        self._store.delete(cache_key)


class CacheFactory(object):

    def __init__(self, store, prefix=None, default_timeout=DEFAULT_TIMEOUT):
        self._store = store
        self._prefix = prefix
        self._default_timeout = default_timeout

    @staticmethod
    def from_store_url(store_type, store_url, prefix=None, default_timeout=DEFAULT_TIMEOUT):
        if store_type == "redis":
            store = RedisStore.from_url(store_url)
            return CacheFactory(store, prefix, default_timeout).cache
        else:
            raise NotImplementedError

    @staticmethod
    def from_store_conn(store_type, conn, prefix=None, default_timeout=DEFAULT_TIMEOUT):
        if store_type == "redis":
            store = RedisStore(conn)
            return CacheFactory(store, prefix, default_timeout).cache
        else:
            raise NotImplementedError

    def cache(self, ruler, timeout=None):
        def wrapper(func):
            return Cache(self._store, self._prefix, func, ruler, timeout or self._default_timeout)
        return wrapper


if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from abc import ABCMeta, abstractmethod
import six

from redis import StrictRedis

from .constants import NOT_FOUND, DEFAULT_TIMEOUT
from .serializer import Serializer


@six.add_metaclass(ABCMeta)
class AbstractStore(object):

    @abstractmethod
    def get(self, key, default=NOT_FOUND):
        pass

    @abstractmethod
    def set(self, key, value, timeout=DEFAULT_TIMEOUT):
        pass

    @abstractmethod
    def delete(self, *keys):
        pass


class RedisStore(AbstractStore):

    def __init__(self, conn, serializer=Serializer(format="JSON")):
        self._conn = conn
        self._serializer = serializer

    @staticmethod
    def from_url(url):
        conn = StrictRedis.from_url(url)
        return RedisStore(conn)

    def get(self, key, default=NOT_FOUND):
        data = self._conn.get(key)
        if data is None:
            return default
        return self._serializer.decode(data)

    def set(self, key, value, timeout=DEFAULT_TIMEOUT):
        data = self._serializer.encode(value)
        self._conn.setex(key, timeout, data)

    def delete(self, *keys):
        self._conn.delete(*keys)



from setuptools import setup, find_packages
from codecs import open
from os import path

__version__ = '0.0.7'

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# get the dependencies and installs
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')

install_requires = [x.strip() for x in all_reqs if x]

setup(
    name='hucache',
    version=__version__,
    description='A Declarative Caching Library for Human',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/wayhome/hucache',
    license='MIT',
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License'
    ],
    keywords='',
    packages=find_packages(exclude=['docs', 'tests*']),
    include_package_data=True,
    author='wayhome',
    install_requires=install_requires,
    author_email='yanckin@gmail.com',

)
